#+
# Name:
#	snpp
# PURPOSE:
#	calculate the S/N per pixel for CSST and simulate a noisy spectrum for any given template.
# CALLING SEQUENCE:
#   snpp,limitmag, repeatnum=10,obstime=300,targetmag=18,/skyperpixel,$
#         galtpl=,wavearr=wavearr,mockgal=mockgal,galflux=galflux
#   plot, wavearr, galflux  ; the input galaxy template
#   plot, wavearr, mockgal  ; the output spectrum with noise
#     
# INPUTS:
# OPTIONAL IUTPUTS:
#   darkcurrent    dark current, in e/s/pix, (defult: 0.0017)
#   deltal         the delta lambda per pixel, in unit of nm (defult: 0.1755555 nm)
#   fovp           diameter of fiber (or spaxel) in arcsec (defult: 0.2 arcsec)
#   filtera         the filter you chosed to estimate the S/N (defult: bessell_V)
#   galtpl         the filename of star-forming galaxy template you want to use. 
#                  They are in the ../obs/SFgal_tpl/ folder (default: SFgal_texp_FeH0_tau5_Ew10.fits)
#   lambdac        the noise at this wavelength wanted (defult: 550 nm)
#   npixel_width   the width of the spectrum on the CCD (defult: 3.0)
#   obstime        in seconds, single integration time (defult: 300s)
#   outfile        the output file name (defult: '../results/noise.dat' )
#   qinput         the throughput correct factor (defult: 1.0)
#   readnoise      read noise, in e/pix. (defult: 4.0)
#   redshift       the redshift of the target spectrum. (defult: 0.0)
#   repeatnum      repeat number (defult: 1.0)
#   skyperpixel    a second way of estimating the Sky, if know the sky photon number per pixel
#   skyv           V band sky brightness in Johnson V mag/arcsec^2 unit (defult: 22.5 mag/arcsec^2)
#  slitwidth      suit to the slit case. the length assumed to be 0.15 arcsec
#   snlimit        S/N limit (defult: 1.0)
#   specsample     pixels per spectral resolution element (defult: 2)
#   targetmag      the surface brightness of the target you want to calculate the S/N (defult: 22 .5 mag/arcsec^2)
#   teld           diameter of the telescope, in cm unit. (defult: d=200 cm)
# OUTPUTS:
#   limitmag       the Vband surface brightness needed to achieve the S/N limit (defult: 1.0)
# OPTIONAL OUTPUTS:
#   limitemi       the medien of Flambda*dlambda*sampling value of Ha line
#   limitemif      the limit detection of Ha flux 
#   snmean         the median S/N of the whole input target spectrum (mag_v=targetmag)
#   wavearr        the wave array (nm)
#   galflux        the input galaxy flux  (1e-13 erg/s/cm2/nm)
#   mockgal        the mocked galaxy flux  with noise (1e-13 erg/s/cm2/nm)
#
# v5: 15 August 2018      writen by Lei Hao, rivised by Jun Yin
# v7: 10 Sep 2019  by Jun Yin
#     1) remove the function im_filtermag, so do not need the Kcorrect package anymore.
#     2) 
#python 
# v7: 22 Sep 2019 by Mengting Ju
#1.02alpha : 26 Sep 2019 
#1.03alpha: 29 Sep 2019
#1.03beta: 14 Oct 2019
#-
#####################################################################################
#####################################################################################

from astropy.io import fits
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import pandas as pd
from scipy import interpolate
from sympy import *
import os
import pdb

####################################################################################
################################################################################
class read_filter(object):
    # load the filters
    # filterfile = filtera  # '../sdss_g0.par'
    # filterpath='./'
    # filterfile=filterpath+filtersel   # ;fluxfilter: max=1, min=0, no particular unit
    def __init__(self, filterfile=''):

        print('Filter File:', filterfile)

        band = pd.read_csv(filterfile, sep='\s+', header=None, comment='#')
        self.wave = band[0].values  # A
        self.throughput = band[1].values  # vaccum_pass
        self.wavemin = self.wave[0]
        self.wavemax = self.wave[-1]

        # find the central wavelength, effective wavelength, and FWHM of the given filter
        filtermid = (self.wavemax - self.wavemin) * 0.5  # A, central wavelength
        dwave = self.wave[1:] - self.wave[:-1]
        self.waveeff = np.nansum(dwave * self.wave[1:] * self.throughput[1:]) / \
                       np.nansum(dwave * self.throughput[1:])
                                                                            # A, effective wavelength
        rmax = np.max(self.throughput)
        nnn = np.where(self.throughput > 0.5 * rmax)[0]
        self.FWHMmin = self.wave[nnn[0]]      # wave range at FWHM
        self.FWHMmax = self.wave[nnn[-1]]
        self.wavefwhm = self.FWHMmax - self.FWHMmin  # A, FWHM

####################################################################################
#select input parameters
def input_mag_model(targetmag, galtpl, filtera):

    #filter
    resulte = read_filter(filterfile=filtera)
    wavefilter = resulte[0]
    fluxfilter = resulte[1]
    vmin = resulte[2]
    vmax = resulte[3]
    filtereff = resulte[4]

    ####################################################################################
    # define wavelength array,
    #cover the range of 350nm to 1050nm, depend on the spectral resolution wanted.

    delta_lambda = 1.755555 # has to be in unit of A
    print('delta_lambda:', delta_lambda)

    narray = int((10000.0-3500.0) / delta_lambda)
    wavearr = 3500.0 + delta_lambda * np.float64(np.arange(narray))
    # select out the array of V band filter
    ii = np.logical_and(wavearr >= vmin, wavearr <= vmax)
    wavetmp2 = wavearr[ii]
    x = np.interp(wavetmp2,wavefilter, fluxfilter)
    integratef4 = x * wavetmp2
    integconst = simps(integratef4, wavetmp2) # int(lambda*Rlambda*dlambda)

    lambdav = filtereff #A
    ###############################################################
    # define basic target brightness, parameters constantly change
    itarget = targetmag    # in Johnson V mag/arcsec^2 unit
    print('itarget:', itarget)

    itarget_jy = 3631.0 * 10**(-itarget/2.5 + 3.0)  # target flux in V in mJy/arcsec^2 unit
    itarget_nm = itarget_jy*3.0 / (lambdav / 100.0)**2 # target flux in V in 10^(-12)erg/s/cm^2/A (/arcsec^2 ?)

    galtpl = galtpl
    tplfile = galtpl
    print('tplfile:',tplfile)

    sfgal = fits.open(tplfile)
    wavegal = sfgal[1].data['wavelength'] # A
    galflux2 = sfgal[1].data['flux']
    galflux1 = np.interp(wavearr, wavegal, galflux2)

    #;normalize the galaxy spectrum to the V band magnitude specified.
    ii = np.logical_and(wavegal >= vmin, wavegal <= vmax)
    wavetmp = wavegal[ii]
    fluxtmp = galflux2[ii]
    x = np.interp(wavetmp, wavefilter, fluxfilter)
    vfluxtmp = x * wavetmp * fluxtmp #bandpass*lambda*F_gal_lambda
    galintegrate = simps(vfluxtmp, wavetmp)
    galnorm = itarget_nm * integconst / galintegrate
    galflux = galnorm * galflux1   # the unit should now be in 10^(-12)erg/s/A/cm^2 (/arcsec^2 ?)

    return wavearr, galflux

#######################################################################################################
# select put in wave and flux
def input_wave_flux(wave,flux):

    wave=wave
    flux=flux
    delta_lambda=1.755555
    narray=int((wave[-1]-wave[0])/delta_lambda)
    wavearr=wave[0]+delta_lambda*np.float64(np.arange(narray))

    galflux=np.interp(wavearr,wave,flux)      #erg/s/A/cm2

    return wavearr, galflux*10**12

###############################################################################
def filter_mag(objwave, objflux, filterwave, filterthroughtput, output='mag'):
    '''
    :param objwave: unit as A
    :param objflux: unit as erg/s/cm^2/A
    :param filterwave: unit as A
    :param filterthroughtput: unit as detector signal per photon (use vaccum_pass for space telescope,
                              otherwise select a given airmass)
    :return: AB magnitude in this band
    '''

    # resample the throughtput to objwave
    ind = (objwave >= np.min(filterwave)) & (objwave <= np.max(filterwave))
    wavetmp = objwave[ind]
    phot_frac = np.interp(wavetmp, filterwave, filterthroughtput)  # phot fraction (/s/A/cm^2?)
    # convert to energy fraction
    # E_frac = E0/hv * N_frac / E0 ~ N_frac/nu ~ N_frac * lambda
    energy_frac = phot_frac * wavetmp  # convert to energy fraction

    # convert the objflux to erg/s/cm^2/Hz
    c = 3e18    # A/s
    objflux_hz = objflux[ind] * (objwave[ind]**2) / c    # erg/s/scm^2/Hz

    from scipy.integrate import simps
    integrate_flux = simps(objflux_hz * energy_frac, c/wavetmp) / simps(energy_frac, c/wavetmp)

    if output == 'mag':
        mag = -2.5 * np.log10(integrate_flux) - 48.6
        return mag
    elif output == 'flux':
        return integrate_flux
    else:
        print('Error: output need to be "mag" or "flux". ')
        return np.nan


def get_snr(darkc=None, repn=None, obst=None, npixw=None, specsampling=None,
            rn=None, nsky=None, isource=None):

    darkn = isource.copy() * 0.0
    darkn = darkn + (darkc * repn * obst * npixw * specsampling)  # e/s/pix * sampling * t * extraction area
    rnn2 = isource.copy() * 0.0
    rnn2 = rnn2 + rn ** 2 * (repn * npixw * specsampling)  # e/pix * sampling * readout times * extraction area
    sourcenn = (isource * repn * obst)  # e/s/sampling * t per spaxel
    skynn = (nsky * repn * obst)  # e/s/sampling * t per spaxel

    noise = np.sqrt(rnn2 + darkn + skynn + sourcenn)  # total noise
    snr = sourcenn / noise  # S/N
    # snarray[i] = sn1
    # nn = np.sqrt(rnn2 + darkn + skynn)  # system noise

    return snr


def get_expt(darkc=None, repn=None, snr=None, npixw=None, specsampling=None,
            rn=None, nsky=None, isource=None):

    coeff_t2 = (isource * repn)**2
    coeff_t1 = -snr**2 * ((darkc * repn * npixw * specsampling) + (nsky * repn) + (isource * repn))
    coeff_t0 = -snr**2 * rn ** 2 * (repn * npixw * specsampling)

    t1 = (-coeff_t1 + np.sqrt(coeff_t1**2 - 4 * coeff_t2 * coeff_t0)) / (2 * coeff_t2)
    t2 = (-coeff_t1 - np.sqrt(coeff_t1**2 - 4 * coeff_t2 * coeff_t0)) / (2 * coeff_t2)

    if max(t1, t2) > 0 and min(t1, t2) < 0:
        return max(t1, t2)
    else:
        print('check the solution of t1, t2!')
        return np.nan


def get_isource(darkc=None, repn=None, snr=None, npixw=None, specsampling=None,
            rn=None, nsky=None, obst=None):

    coeff_s2 = (repn * obst)**2
    coeff_s1 = -snr**2 * (repn * obst)
    coeff_s0 = -snr**2 * (rn ** 2 * (repn * npixw * specsampling) +
                          (darkc * repn * obst * npixw * specsampling) +
                          (nsky * repn * obst))

    s1 = (-coeff_s1 + np.sqrt(coeff_s1**2 - 4 * coeff_s2 * coeff_s0)) / (2 * coeff_s2)
    s2 = (-coeff_s1 - np.sqrt(coeff_s1**2 - 4 * coeff_s2 * coeff_s0)) / (2 * coeff_s2)

    if max(s1, s2) > 0 and min(s1, s2) < 0:
        return max(s1, s2)
    else:
        print('check the solution of s1, s2!')
        return np.nan


def flux2electron(wavearr, fluxarr, throughtputwave, throughtputvalue, fluxunit=''):
    '''
    :param wavearr: A
    :param fluxarr: erg/s/cm^2/A
    :return:
    '''

    cc = 3.0e18   # speed of light, A/s
    spaxel_area = 0.2**2    # arcsec^2
    d = 200  # diameter of the telescope, in cm unit
    obscure = 0.0  # effective central obscuration, no unit
    telarea = 3.14159 / 4.0 * d * d * (1.0 - obscure)  # effective area of the telescope, cm^2
    delta_lambda_per_pixel = 1.755555  # per pixel
    delta_hz_per_specpixel = delta_lambda_per_pixel / wavearr**2 * cc  # s^-1
    specsampling = 2.0  # 2 pixels
    planckh = 6.626e-27  # erg*s
    hv = planckh * cc / wavearr  # erg
    QE = 1

    qvalue = np.interp(wavearr, throughtputwave, throughtputvalue)

    if fluxunit=='erg/s/cm^2/A':
        isource = fluxarr * wavearr ** 2 / cc * spaxel_area  # erg/s/cm^2/Hz per spaxel
        isource2 = isource * telarea  # erg/s/Hz per spaxel
        isource3 = qvalue * isource2 * delta_hz_per_specpixel  # erg/s per spaxel
        isource4 = isource3 * specsampling  # erg/s per spec-element per spaxel
        isource5 = isource4 / hv  # phot/s per spec-element per spaxel
        isource6 = isource5 * QE  # e/s per spec-element per spaxel

        return isource6
    else:

        return np.nan


####################################################################################
class snpp(object):

    def __init__(self, wavearr=False, galflux=False,
                 calculate_type=1, readnoise=4.0,
                 fovp=0.2, npixel_width=2.0, snlimit=10, targetmaglimit=None,
                 obstime=300, repeatnum=1.0, skyr=22.5, qinput=1.0,
                 skyperpixel=False, bb=0.):

        # wavearr need to be in unit as A, delta_wavearr need to be 1.755A
        # galflux need to be in unit as erg/s/cm^2/A/arcsec^2

        # Do extensive checking of possible input errors
        #
         
        self.qinput = qinput
        self.fovp = fovp
        self.obstime = obstime
        self.skyr = skyr
        self.repeatnum = repeatnum
        self.readnoise = readnoise
        self.npixel_width = npixel_width
        self.skyperpixel = skyperpixel
        # self.filename = filename
        ###########################################################################
        #some basic unchanged parameters     
        d = 200        # diameter of the telescope, in cm unit
        obscure = 0.0  #effective central obscuration, no unit
        telarea = 3.14159 / 4.0 * d * d * (1.0 - obscure)  #effective area of the telescope, cm^2
        darkc = 0.017   #dark current, in e/s/pix
        planckh = 6.626e-27    # erg*s
        cc = 3.0e18   # speed of light, A/s
        slitunit = 0.074  # arcsec. pixelscale on 主巡天 CCD
        
        rn = readnoise  #read noise, in e/pix
        print('readnoise:', rn)
                       
        npixw = npixel_width
        print('npixel_width:', npixw)
               
        obst = obstime  # in seconds, single integration time
        print('obstime:', obst)
        
        repn = repeatnum   # repeating time
        print('repeatnum:', repn)

        iskyr0 = skyr  # in Johnson V mag/arcsec^2 unit
        print('skyv:', iskyr0)
        
        specsampling = 2.0
        #print('specsample:', sampling)
        
        delta_lambda_per_pixel = 1.755555     # per pixel
        
        #####################################################################

        # some less basic parameters, may change, but not often
        
        throughput = pd.read_csv(os.getenv('SNPP_PATH') + 'obs/IFU_throughput.dat',
                               sep='\s+', skiprows=1, header=None, names=['nm','q'])
        lambdaq = throughput.nm.values * 10    # nm to A
        qtot = throughput.q.values             #; throughput of the whole system,
 

        #;assuming the total throughput cannot reach the theory value, 0.3 is the upper limit. 
        qtot[qtot>=0.3] = 0.3
        
        qinput = qinput
        print('qinput:', qinput)
        
        tot_efficiency = qtot * qinput #*qe ;qtot of CSST already includes the CCD efficiency
        

        spaxel_area = (fovp)**2  # 1 spaxel in IFU, arcsec^2
        print('spaxel_area:', spaxel_area)
        ##############################################################################
        
        # SKY
 
        if self.skyperpixel==True :
            # print(1)
            # since the numbers are given by the main survey,
            # our detected Sky electron will be less, so scale a rough factor of 0.9
            fluxskypp = np.zeros(len(wavearr))
            scaletemp = 0.9

            ii = np.logical_and(wavearr >= 2550, wavearr <= 4000)
            counta = len(np.where(ii==1)[0])
            if counta > 0:
                fluxskypp[ii] = 0.028/counta     # e/s/spaxel/pixel
            ii = np.logical_and(wavearr >= 4000, wavearr <= 6000)
            countb = len(np.where(ii==1)[0])
            if countb > 0:
                fluxskypp[ii] = 0.229/countb
            ii = np.logical_and(wavearr >= 6000, wavearr <= 9000)
            countc = len(np.where(ii==1)[0])
            if countc > 0:
                fluxskypp[ii] = 0.301/countc
            ii = np.where(wavearr >= 9000)[0]
            countd = len(ii)
            if countd > 0:
                fluxskypp[ii] = 0.301/countd

            fluxskypp = fluxskypp / slitunit**2 * spaxel_area * scaletemp  # e/s/spaxel/spec-elements

        else:
            # print(2)
            filter = read_filter(os.getenv('SNPP_PATH')+'obs/filters/sdss_r0.par')
            # wavefilter = resulte[0]  # A
            # fluxfilter = resulte[1]  # vaccum_pass per photon
            # vmin = resulte[2]
            # vmax = resulte[3]
            # filtereff = resulte[4]
            # FWHMmin = resulte[5]
            # FWHMmax = resulte[6]

            # select out the array of r band filter
            ii = np.logical_and(wavearr >= filter.wavemin, wavearr <= filter.wavemax)
            wavetmp2 = wavearr[ii]
            x = np.interp(wavetmp2, filter.wave, filter.throughput)     # phot fraction (/s/A/cm^2?)
            integratef4 = x * wavetmp2      # convert to energy fraction
            integconst = simps(integratef4, wavetmp2)   # int(lambda*Rlambda*dlambda)

            #####################################################################

            #define r band sky brightness

            lambdar = filter.waveeff   #in A

            #sky brightness corresponding to this sky magnitude
            iskyr0_mjy = 3631.0 * 10**(-iskyr0 / 2.5 + 3.0)  # sky flux in V in mJy/arcsec^2 unit
            iskyr0_A = iskyr0_mjy * 3.0 / (lambdar/100.0)**2   # sky flux in V in 10^(-12)erg/s/cm^2/A/arcsec^2
                                                            # sky flux in V in 10^(-12)erg/s/cm^2/nm/arcsec^2

            #readin the ground sky spectrum 
            skybg_50 = pd.read_csv(os.getenv('SNPP_PATH')+'obs/skybg_50_10.dat',sep='\s+',header=None,skiprows=14)
            wavesky = np.array(skybg_50[0]) * 10 # in A
            fluxsky1 = np.array(skybg_50[1]) / 10 # phot/s/A/arcsec^2/m^2
            fluxsky2 = fluxsky1 / wavesky * 1.98 # change the sky flux unit to 10^(-12)erg/s/cm^2/A/arcsec^2


            # This fluxsky is in unit of phot/s/A/arcsec^2/m^2, to convert it to F_lambda/arcsec^2,
            # need to do fluxsky(phot/s/A/arcsec^2/m^2) * h(6.625*10^{-27}erg.s) * nu(1/s) * 10{-4}(m^2/cm^2)
            # = fluxsky * c(3.0*10^{18}A/s)/lambda(A)*6.6*10{-31} erg/s/cm^2/A/arcsec^2
            # = fluxsky / lambda * 1.98 * 10^{-12} erg/s/cm^2/A/arcsec^2

            # find out the normalization of the sky,
            ii = np.logical_and(wavesky >= vmin, wavesky <= vmax)
            wavetmp = wavesky[ii]
            fluxtmp = fluxsky1[ii]  # phot/s/A/arcsec^2/m

            x = np.interp(wavetmp, wavefilter, fluxfilter)
            vfluxtmp = x * fluxtmp * 1.98                 # 10^(-12) erg/s/cm^2/A/arcsec^2 * A
            skyintegrate = simps(vfluxtmp, wavetmp)
            skynorm = iskyr0_A * integconst / skyintegrate
            fluxsky3 = np.interp(wavearr, wavesky, fluxsky2)
            fluxsky = fluxsky3 * skynorm
            # get the sky spectrum in wavearr grid, the unit should now be the same as fluxvega:
            # 10^(-12) erg/s/A/cm^2  (/arcsec^2 ?)

            fluxskypp = fluxsky


        ##########################################################################
        
        #define observation information, parameters constantly change
       
        nwave = len(wavearr)   # per delta_wavearr
        expf2 = np.zeros(nwave)
        QE = 1

        # 'lambda', 'S/N', 'tar_flux', 'tot_noise', 'sc_noise', \
        # 'sys_noise', 'readnoise', 'dark_noise', 'sky_noise', 'mockgal'
        snarray = np.zeros(nwave)
        tot_noise = np.zeros(nwave)
        sc_noise = np.zeros(nwave)
        sys_noise = np.zeros(nwave)
        read_noise = np.zeros(nwave)
        dark_noise = np.zeros(nwave)
        sky_noise = np.zeros(nwave)
        mockflux = np.zeros(nwave)
        mockerror = np.zeros(nwave)
        mockgal = np.zeros(nwave)
        isource6 = np.zeros(nwave)

        tmp = np.zeros(nwave)
        lista = np.zeros(shape=(nwave, 10), dtype=float)


        isource = flux2electron(wavearr, galflux, lambdaq, tot_efficiency, fluxunit='erg/s/cm^2/A')    # e/s
        if self.skyperpixel:
            nsky = fluxskypp * specsampling  # e/s in npixw * sampling pixels


        snr = get_snr(darkc=darkc, repn=repn, obst=obst, npixw=npixw, specsampling=specsampling,
                      rn=rn, nsky=nsky, isource=isource)



        pdb.set_trace()



        for i in range(nwave):     # for each spec-pixel

            ilambda = wavearr[i]
            iqlambda = np.interp(ilambda, lambdaq, tot_efficiency)
            hv = planckh * cc / ilambda     # erg
            delta_hz_per_specpixel = cc * delta_lambda_per_pixel / ilambda / ilambda    # s^-1
            delta_hz_per_specelement = delta_hz_per_specpixel * specsampling
            
            # now that many fluxes are in 10^(-12)erg/s/A/cm^2, to convert it to Jy, need to multiple:
            # lambda0^2/c(in A) = lambda0^2(A)/(3.*10^(18))*10^(-12)erg/s/Hz/cm^2
            # = lambda^2(A)*3.33*10^(-31)erg/s/Hz/cm^2=lambda^2(A)*3.33*10^(-8)Jy
            # = lambda^2(A)*0.0333uJy

            if self.skyperpixel :
                nsky = fluxskypp[i] * specsampling    # e/s in npixw * sampling pixels
            else:
                #find out sky value at lambda0
                #calculate n_sky/pixel
                isky = fluxskypp[i] * lambda0**2 * 0.0333 * spaxel_area   #in uJy/spaxel unit
                iskyall = isky * telarea / 1000.0   #in 10-26 erg/s/Hz /spaxel
                fsky = qlambda * iskyall * delta_hz   #10^{-8} erg/s /spaxel
                fsky = fsky * specsampling
                nsky = fsky / hv * 10.0   #in unit of #e/s /spaxel

            #calculate n_source/pixel
            # isource = galflux[i] * lambda0**2 * 0.0333 * spaxel_area   #in uJy/spaxel unit
            # isall = isource * telarea / 1000.0   #in 10-26 erg/s/Hz /spaxel
            # fs = qlambda * isall * delta_hz   #10^{-8} erg/s /spaxel
            # fs = fs * specsampling
            # ns = fs / hv * 10.0         # photon/s/spaxel, need to consider QE curve
            #                             # in unit of #e/s /spaxel

            isource = galflux[i] * ilambda**2/cc * spaxel_area          # erg/s/cm^2/Hz per spaxel
            isource2 = isource * telarea                                # erg/s/Hz per spaxel
            isource3 = iqlambda * isource2 * delta_hz_per_specpixel     # erg/s per spaxel
            isource4 = isource3 * specsampling                          # erg/s per spec-element per spaxel
            isource5 = isource4 / hv                                    # phot/s per spec-element per spaxel
            isource6[i] = isource5 * QE                                    # e/s per spec-element per spaxel

            # need to consider additional fraction if extracting area cannot include total source flux
            # (PSF size larger than npixw)
            # (??) the result will be same for extended source? flux in other spaxels will be counted in.


            isource = flux2electron(wavearr, galflux, lambdaq, tot_efficiency, fluxunit='erg/s/cm^2/Hz')
            pdb.set_trace()


            darkn = (darkc * repn * obst * npixw * specsampling)        # e/s/pix * sampling * t * extraction area
            rnn2 = rn**2 * (repn * npixw * specsampling)                # e/pix * sampling * readout times * extraction area
            sourcenn = (isource6[i] * repn * obst)                         # e/s/sampling * t per spaxel
            skynn = (nsky * repn * obst)                                # e/s/sampling * t per spaxel


            nn1 = np.sqrt(rnn2 + darkn + skynn + sourcenn)  # total noise
            sn1 = sourcenn / nn1  # S/N
            snarray[i] = sn1
            nn = np.sqrt(rnn2 + darkn + skynn)  # system noise
            
            
            # mockgal[i] = galflux[i] + galflux[i] / snarray[i] * np.random.randn(1,1)[0][0]  # in 10^{-12} erg/s/A/cm^2
            mockflux[i] = np.random.normal(galflux[i], scale=galflux[i]/snarray[i])
            mockerror[i] = galflux[i] / snarray[i]

            lista[i,:] = [ilambda, sn1, galflux[i], nn1,\
                        np.sqrt(sourcenn), nn, np.sqrt(rnn2), np.sqrt(darkn), \
                        np.sqrt(skynn), mockflux[i]]



            snr = get_snr(darkc=darkc, repn=repn, obst=obst, npixw=npixw, specsampling=specsampling,
                     rn=rn, nsky=nsky, isource=isource6[i])
            t = get_expt(darkc=darkc, repn=repn, snr=snarray[i], npixw=npixw, specsampling=specsampling,
                     rn=rn, nsky=nsky, isource=isource6[i])
            s = get_isource(darkc=darkc, repn=repn, snr=snarray[i], npixw=npixw, specsampling=specsampling,
                     rn=rn, nsky=nsky, obst=obst)   # identical to the f1 in the old version.



            # pdb.set_trace()


            # set the detection limit
            # detlimit = snlimit
            detlimit = snarray[i]

            nntmp = detlimit**2 + np.sqrt(detlimit**4 + 4.0 * detlimit**2 * nn**2)
            nntmp = nntmp / 2.0
            # calculate detection limit in uJy and mag
            fnn = nntmp
            f1 = fnn / obst / repn    #in e/s
            f2 = f1 * hv       #in 10^{-9} erg/s
            f3 = f2 / delta_lambda_per_pixel #in 10^{-9} erg/s/A
            f1 = f3 / telarea   #in 10^{-9} erg/s/A/cm^2
            f2 = f1 / iqlambda    #in 10^{-9} erg/s/A/cm^2
            expf2[i] = f2 / spaxel_area * 100000.   # in 10^{-14} erg/s/A/cm^2/arcsec^2


        pdb.set_trace()

        ############################################################################################
        self.output_spectra = lista
        # 'lambda', 'S/N', 'tar_flux', 'tot_noise', 'sc_noise', \
        # 'sys_noise', 'readnoise', 'dark_noise', 'sky_noise', 'mockgal'


        #####################################################################
        filter = read_filter(filterfile=os.getenv('SNPP_PATH')+'obs/filters/sdss_g0.par')
        # wavefilter = resulte[0]
        # fluxfilter = resulte[1]
        # vmin = resulte[2]
        # vmax = resulte[3]
        # filtereff = resulte[4]
        # FWHMmin = resulte[5]
        # FWHMmax = resulte[6]

        ii = np.logical_and(wavearr >= filter.FWHMmin , wavearr <= filter.FWHMmax)
        wavetmp = wavearr[ii]
        snm = np.median(snarray[ii])
        # the median SN of FWHM range to achieve the sn limit
        im = np.where(snarray[ii] == snm)[0]
        fact = expf2[ii][im] * 0.01 / galflux[ii][im]
        limitmag = -2.5 * np.log10(fact) + targetmaglimit
        self.aa = limitmag
        print('limitmag:', limitmag)

        # pdb.set_trace()


        if calculate_type == 1:
            # signal_at_targetband = filter_mag(wavearr, mockflux, filter.wave, filter.throughput, output='flux')
            # noise_at_targetband = filter_mag(wavearr, mockerror, filter.wave, filter.throughput, output='flux')
            ii = np.logical_and(wavearr >= filter.FWHMmin, wavearr <= filter.FWHMmax)
            median_snr = np.median(snarray[ii])
            self.result = median_snr

        elif calculate_type == 2:
            self.result = exptime_need

        elif calculate_type == 3:
            self.result = surface_brightness

        else:
            self.result = np.nan



    ############################################################################

    # def printf(self):
    #     return self.aa

    #############################################################################

    # def fits(self):
    #     return self.bb

    ###############################################################################

    def save_output_spectra(self, outfilename=''):

        list = self.output_spectra
        namedat = np.array(['lambda', 'S/N', 'tar_flux', 'tot_noise', 'sc_noise', \
                          'sys_noise', 'readnoise', 'dark_noise', 'sky_noise', 'mockgal'])
        unit = np.array(['A', ' ','1e-12 erg/s/cm2/A',\
                       '#e','#e','#e','#e','#e','#e', '1e-12 erg/s/cm2/A'])

        hdr = fits.Header()
        for i in range(len(namedat)):
            hdr[str(i)] = unit[i]

        hdu0 = fits.PrimaryHDU(header=hdr)
        hdu1 = fits.BinTableHDU.from_columns(
            [fits.Column(name=namedat[i], array=np.array(list[:,i]), format='1E') for i in range(len(namedat))]
        )
        hdulist = fits.HDUList([hdu0, hdu1])

        print('output filt:',outfilename)
        hdulist.writeto(outfilename, overwrite=True)



    # print('The END!')
        
#######################################################################################################




import numpy as np
import pandas as pd
import os
import astropy.io.fits as fits
from scipy.integrate import simps
import pdb

refdata = os.path.abspath(os.path.dirname(__file__)+'/../refdata') + '/'


class read_template(object):

    def __init__(self, source):

        self.source = source
        self.line_definitions = self.source['lines']

        # snpp_path = '/Users/linlin/Pro/snpp_v1_05/refdata/'
        template_filename = refdata + 'sed/' + self.source['name']
        # filtera = snpp_path + 'normalization/filters/sdss_g0.par'
        # result = input_mag_model(self.source['normalization']['value'],
        #                          galtpl, filtera)  # scale the template to given brightness
        hdu = fits.open(template_filename)
        self.wave = hdu[1].data['wavelength']   # A
        self.flux = hdu[1].data['flux'] * 1e-12         # 10^-12 erg/s/A/cm2


class read_filter(object):
    # load the filters
    # filterfile = filtera  # '../sdss_g0.par'
    # filterpath='./'
    # filterfile=filterpath+filtersel   # ;fluxfilter: max=1, min=0, no particular unit
    def __init__(self, band):

        filterfile = refdata + 'normalization/filters/sdss_g0.par'
        # print('Filter File:', filterfile)

        band = pd.read_csv(filterfile, sep='\s+', header=None, comment='#')
        self.wave = band[0].values  # A
        self.throughput = band[1].values  # vaccum_pass
        self.wavemin = self.wave[0]
        self.wavemax = self.wave[-1]

        # find the central wavelength, effective wavelength, and FWHM of the given filter
        filtermid = (self.wavemax - self.wavemin) * 0.5  # A, central wavelength
        dwave = self.wave[1:] - self.wave[:-1]
        self.waveeff = np.nansum(dwave * self.wave[1:] * self.throughput[1:]) / \
                       np.nansum(dwave * self.throughput[1:])
                                                                            # A, effective wavelength
        rmax = np.max(self.throughput)
        nnn = np.where(self.throughput > 0.5 * rmax)[0]
        self.FWHMmin = self.wave[nnn[0]]      # wave range at FWHM
        self.FWHMmax = self.wave[nnn[-1]]
        self.wavefwhm = self.FWHMmax - self.FWHMmin  # A, FWHM


def filter_mag(objwave, objflux, filterwave, filterthroughtput, output='mag'):
    '''
    :param objwave: unit as A
    :param objflux: unit as erg/s/cm^2/A
    :param filterwave: unit as A
    :param filterthroughtput: unit as detector signal per photon (use vaccum_pass for space telescope,
                              otherwise select a given airmass)
    :return: AB magnitude in this band
    '''

    # resample the throughtput to objwave
    ind = (objwave >= np.min(filterwave)) & (objwave <= np.max(filterwave))
    wavetmp = objwave[ind]
    phot_frac = np.interp(wavetmp, filterwave, filterthroughtput)  # phot fraction (/s/A/cm^2?)
    # convert to energy fraction
    # E_frac = E0/hv * N_frac / E0 ~ N_frac/nu ~ N_frac * lambda
    energy_frac = phot_frac * wavetmp  # convert to energy fraction

    # convert the objflux to erg/s/cm^2/Hz
    c = 3e18    # A/s
    objflux_hz = objflux[ind] * (objwave[ind]**2) / c    # erg/s/scm^2/Hz

    from scipy.integrate import simps
    integrate_flux = simps(objflux_hz * energy_frac, c/wavetmp) / simps(energy_frac, c/wavetmp)

    if output == 'mag':
        mag = -2.5 * np.log10(integrate_flux) - 48.6
        return mag
    elif output == 'flux':
        return integrate_flux
    else:
        print('Error: output need to be "mag" or "flux". ')
        return np.nan


def normalized(template_wave, template_flux, config):

    # grid = get_grid(config)

    source = config['source']

    normalize = source['normalization']

    filter = read_filter(normalize['band'])

    ind_filter = (template_wave >= filter.wavemin) & (template_wave <= filter.wavemax)
    filter_wave = template_wave[ind_filter]
    filter_flux = np.interp(filter_wave, filter.wave, filter.throughput)
    filter_constant = simps(filter_flux * filter_wave, filter_wave)

    template_constant = simps(filter_flux * template_wave[ind_filter] * template_flux[ind_filter],
                              template_wave[ind_filter])

    u0 = normalize['value']
    u0 = 10**((u0 + 48.6)/(-2.5))         # target flux in erg/s/cm^2/Hz unit
    u0 = u0 * 3e18 / filter.waveeff**2    # erg/s/cm^2/A

    # geometry = source['geometry']
    # if geometry['type'] == 'Sersic':
    #
    #     xcen = grid.nx / 2 + geometry['x_offset']
    #     ycen = grid.ny / 2 + geometry['y_offset']
    #     rmaj = geometry['Re_major'] / grid.xspxsize
    #     rmin = geometry['Re_minor'] / grid.xspxsize
    #
    #     scaled_img = sersic(grid.nx, grid.ny, xcen, ycen,
    #              u0, rmaj, rmin/rmaj, geometry['PA'], geometry['sersic_n'])
    #
    # elif geometry['type'] == 'flat':
    #
    #     xcen = grid.nx / 2 + geometry['x_offset']
    #     ycen = grid.ny / 2 + geometry['y_offset']
    #
    #     scaled_img = flatbox(grid.nx, grid.ny, xcen, ycen,
    #                     u0, geometry['xwidth'], geometry['ywidth'])
    #
    # else:
    #     print('geometry should be one of these: ("Sersic" | "flat").')
    #     pdb.set_trace()

    factor = u0 * filter_constant / template_constant

    # norm_wave = np.zeros(shape=(len(template_wave), grid.ny, grid.nx), dtype=float)
    # norm_flux = np.zeros(shape=(len(template_wave), grid.ny, grid.nx), dtype=float)
    # for j in range(grid.ny):
    #     for i in range(grid.nx):
    norm_wave = template_wave                  # A
    norm_flux = template_flux * factor         # erg/s/cm^2/A


    # check the normalized magnitude should be normalized.value - 2.5*log(0.04)
    # mag2d = np.zeros(shape=(grid.ny, grid.nx), dtype=float)
    # for j in range(grid.ny):
    #     for i in range(grid.nx):
    mag = filter_mag(norm_wave, norm_flux,
                     filter_wave, filter_flux, output='mag')


    return norm_wave, norm_flux, mag


class ModelCube(object):

    def __init__(self, config):

        self.ccdspec_wave = np.arange(3500, 10000, 1.75555)
        self.ccdspec_nw = len(self.ccdspec_wave)

        # grid = get_grid(config)

        object = config['source']['spectrum']

        template = read_template(object)
        template_wave_interp = self.ccdspec_wave
        template_flux_interp = np.interp(template_wave_interp, template.wave, template.flux)

        # self.ccdspec_flux = np.zeros(shape=(self.ccdspec_nw, grid.ny, grid.nx), dtype=np.float32)

        self.wavecube, self.fluxcube, self.mag2d = normalized(template_wave_interp, template_flux_interp, config)




import os
import json

refdata = os.path.abspath(os.path.dirname(__file__)+'/../refdata') + '/'

def get_telescope_config():

    dict = {}
    dict['diameter'] = 200
    dict['obscure'] = 0.
    dict['coll_area'] = 31415.926535897932

    # out_file = open(refdata + 'csst/telescope/config.json', "w")
    # json.dump(dict, out_file, indent=2)
    # out_file.close()

    return dict


def get_instrument_config():

    dict = {}
    dict['fov_xsize'] = 6                   # arcsec
    dict['fov_ysize'] = 6                   # arcsec
    dict['spaxel_xsize'] = 0.2              # arcsec
    dict['spaxel_ysize'] = 0.2              # arcsec
    dict['ccd_xsize'] = 1.755555            # A, delta_lambda_per_pixel
    dict['ccd_ysize'] = 0.1                 # arcsec, spatial axis
    dict['readout_xbin'] = 1
    dict['readout_ybin'] = 1
    dict['gain'] = 1                            # e/ADU

    dict['dark'] = 0.017                        # e/s/pix
    dict['readout_noise'] = 4.0                 # e/pix
    dict['QE'] = 1.0

    dict['efficiency_file'] = 'IFU_throughput.dat'

    # out_file = open(refdata + "csst/ifs/config.json", "w")
    # json.dump(dict, out_file, indent=2)
    # out_file.close()

    return dict


def build_default_source():

    dict = {}
    dict['spectrum'] = {'name': 'SFgal_texp_FeH0_tau5_Ew10_AGN1.fits', 'redshift': 0.0, 'lines':[]}
    dict['normalization'] = {'value': 17.7, 'unit':'mag/arcsec^2', 'band': 'SDSS-g'}

    # out_file = open(refdata + "source/config.json", "w")
    # json.dump(dict, out_file, indent=2)
    # out_file.close()

    return dict



def build_default_calc():

    calc = {}
    calc['configuration'] = get_instrument_config()
    calc['source'] = build_default_source()
    calc['background'] = 'from_msc'
    calc['background_level'] = 'from_msc'
    calc['repn'] = 20.
    calc['obst'] = 300.

    calc['targetsnr'] = 10      # only be used in calculate_type = 'limitmag',
                                # in this case, normalization value is invalid


    return calc


import pdb
import pandas as pd
import numpy as np
import os
import h5py
from einops import repeat
import matplotlib.pyplot as plt
from . import source
from .source import filter_mag, read_filter
from .config import get_throughput

snpp_refdata = os.path.abspath(os.path.dirname(__file__)+'/../refdata') + '/'

def flux2electron(spaxel_xsize, spaxel_ysize, wavearr, fluxarr, throughtputwave, throughtputvalue, fluxunit=''):
    '''
    :param wavearr: A
    :param fluxarr: erg/s/cm^2/A/arcsec^2
    :return:
    '''

    # instru = config['configuration']

    cc = 3.0e18   # speed of light, A/s
    spaxel_area = spaxel_xsize * spaxel_ysize     # arcsec^2
    d = 200  # diameter of the telescope, in cm unit
    obscure = 0.0  # effective central obscuration, no unit
    telarea = 3.14159 / 4.0 * d * d * (1.0 - obscure)  # effective area of the telescope, cm^2
    delta_lambda_per_pixel = 1.755555  # per pixel
    delta_hz_per_specpixel = delta_lambda_per_pixel / wavearr**2 * cc  # s^-1
    specsampling = 1.0  # 2 pixels
    planckh = 6.626e-27  # erg*s
    hv = planckh * cc / wavearr  # erg
    QE = 1

    if wavearr.ndim == 3:
        qvalue = np.zeros(shape=wavearr.shape, dtype=float)
        qvalue = np.interp(wavearr[:, 0, 0], throughtputwave, throughtputvalue)
        qvalue = repeat(qvalue, 'h -> h '+str(wavearr.shape[1])+' '+str(wavearr.shape[2]))
    elif wavearr.ndim == 1:
        qvalue = np.interp(wavearr, throughtputwave, throughtputvalue)
    else:
        raise ValueError('check dimension of the input array.')

    if fluxunit=='erg/s/cm^2/A':
        isource1 = fluxarr * wavearr ** 2 / cc * spaxel_area  # erg/s/cm^2/Hz per spaxel
        isource2 = isource1 * telarea  # erg/s/Hz per spaxel
        isource3 = qvalue * isource2 * delta_hz_per_specpixel  # erg/s per spaxel
        isource4 = isource3 * specsampling  # erg/s per spec-element per spaxel
        isource5 = isource4 / hv  # phot/s per spec-element per spaxel
        isource6 = isource5 * QE  # e/s per spec-element per spaxel

        return isource6
    else:

        return np.nan


def electron2flux(config, wavearr, countsarr, throughtputwave, throughtputvalue, fluxunit=''):
    '''
    :param wavearr: A
    :param countsarr: e-/s
    :return:
    '''

    instru = config['configuration']

    cc = 3.0e18   # speed of light, A/s
    spaxel_area = instru['spaxel_xsize'] * instru['spaxel_ysize']    # arcsec^2
    d = 200  # diameter of the telescope, in cm unit
    obscure = 0.0  # effective central obscuration, no unit
    telarea = 3.14159 / 4.0 * d * d * (1.0 - obscure)  # effective area of the telescope, cm^2
    delta_lambda_per_pixel = 1.755555  # per pixel
    delta_hz_per_specpixel = delta_lambda_per_pixel / wavearr**2 * cc  # s^-1
    specsampling = 1.0  # 2 pixels
    planckh = 6.626e-27  # erg*s
    hv = planckh * cc / wavearr  # erg
    QE = 1

    qvalue = np.interp(wavearr, throughtputwave, throughtputvalue)


    if fluxunit=='erg/s/cm^2/A':

        isource5 = countsarr / QE  # phot/s per spec-element per spaxel
        isource4 = isource5 * hv  # erg/s per spec-element per spaxel
        isource3 = isource4 / specsampling  # erg/s per spec-element per spaxel
        isource2 = isource3 / delta_hz_per_specpixel / qvalue  # erg/s per spaxel
        isource1 = isource2 / telarea  # erg/s/Hz per spaxel
        fluxarr = isource1 / wavearr ** 2 * cc / spaxel_area  # erg/s/cm^2/Hz per spaxel

        return fluxarr

    else:

        return np.nan


def get_ms_nsky(wavearr, spaxel_xsize, spaxel_ysize):

    # know the sky counts in MS ccd per pixel in different bands,
    # Thus convert it to sky spectrum (counts) in IFS ccd per spaxel

    fluxskypp = np.zeros(len(wavearr))

    ii = np.logical_and(wavearr >= 2550, wavearr <= 4000)
    counta = len(np.where(ii == 1)[0])
    if counta > 0:
        fluxskypp[ii] = 0.028 / counta      # e/s/spaxel/pixel
    ii = np.logical_and(wavearr >= 4000, wavearr <= 6000)
    countb = len(np.where(ii == 1)[0])
    if countb > 0:
        fluxskypp[ii] = 0.229 / countb
    ii = np.logical_and(wavearr >= 6000, wavearr <= 9000)
    countc = len(np.where(ii == 1)[0])
    if countc > 0:
        fluxskypp[ii] = 0.301 / countc
    ii = np.where(wavearr >= 9000)[0]
    countd = len(ii)
    if countd > 0:
        fluxskypp[ii] = 0.301 / countd

    ms_pix_area = 0.074**2      # arcsec
    spaxel_area = spaxel_xsize * spaxel_ysize   # arcsec
    factor = 0.9

    fluxskypp = fluxskypp / ms_pix_area * spaxel_area * factor  # e/s/spaxel/spec-elements

    return fluxskypp


def get_background(background_type, background_level, wavearr, spaxel_xsize, spaxel_ysize):

    # output: sky counts (e-/s), in 1d array.

    if background_type == 'from_msc':
        nsky = get_ms_nsky(wavearr, spaxel_xsize, spaxel_ysize)

    elif background_type == 'zodi':

        # get IFS throughtput
        throughtputwave, throughtputvalue = get_throughput()

        # get the zodi spectrum from Leinert98, convert to e/s/spaxel/spec-elements
        cat = pd.read_csv(os.getenv('SNPP_PATH') + 'refdata/csst/background/zodi_spec_L98.csv', comment='#')
        zodi_wave = cat['wave']
        if background_level in ['high', 'med', 'low']:
            zodi_spec = cat['zodi_' + background_level] / 4.25452e10  # erg/s/cm^2/A/arcsec^2
            zodi_spec_interp = np.interp(wavearr, zodi_wave, zodi_spec)
            zodi_counts_interp = flux2electron(spaxel_xsize, spaxel_ysize,
                                               wavearr, zodi_spec_interp,
                                               throughtputwave, throughtputvalue,
                                               fluxunit='erg/s/cm^2/A')
        else:
            raise ValueError("background_type need to be '(high | med | low). ")

        nsky = zodi_counts_interp

    else:
        raise TypeError("background_type need to be 'from_msc' or 'zodi'. ")


    return nsky



class DetectorCountsRate(source.ModelCube):

    def __init__(self, config):

        self.config = config

        # pdb.set_trace()

        source.ModelCube.__init__(
            self, config
        )

        throughtputwave, throughtputvalue = get_throughput()

        instru = config['configuration']
        self.source_counts = flux2electron(instru['spaxel_xsize'], instru['spaxel_ysize'],
                                           self.wavecube, self.fluxcube,
                                           throughtputwave, throughtputvalue,
                                           fluxunit='erg/s/cm^2/A')
        self.wavearr = self.wavecube[:, 0, 0]



class DetectorCounts(object):

    def __init__(self, config, signal):

        instru = config['configuration']
        ccdspec_wave = signal.ccdspec_wave
        nw = len(signal.ccdspec_wave)
        nspaxel_x = signal.source_counts.shape[1]
        nspaxel_y = signal.source_counts.shape[2]

        # several count_rates
        nsource_rate_cube = signal.source_counts     # e/s

        ndark_rate = np.zeros(shape=nw, dtype=float) + instru['dark']   # e/s/pix

        nreadout_rate = np.zeros(shape=nw, dtype=float) + instru['readout_noise']**2    # e/pix

        nsky_rate = get_background(config['background_type'], config['background_level'],
                                 signal.ccdspec_wave, instru['spaxel_xsize'], instru['spaxel_ysize'])

        # read the exptime
        repn = config['repn']
        obst = config['obst']
        npixw = 2
        specsampling = 1

        # calculate the counts with exptime and extract_area.
        tot_t = repn * obst
        readout_n = repn
        ndark = (ndark_rate * tot_t * npixw * specsampling)  # e/s/pix * sampling * t * extraction area
        nreadout = nreadout_rate * (repn * npixw * specsampling)  # e/pix * sampling * readout times * extraction area
        nsource_cube = (nsource_rate_cube * tot_t)  # e/s/sampling * t per spaxel
        nsky = (nsky_rate * tot_t)  # e/s/sampling * t per spaxel

        # add 2% of the source counts as straylight, scatter into the entire CCD (b: 4k*2k, r: 6k*3k)
        stray_frac = 0.02
        ind_blue = ccdspec_wave < 5700  # assume the blue/red separate at 5700A
        ind_red = ccdspec_wave > 5700
        tot_source_counts_blue = np.sum(nsource_cube[ind_blue, :, :]) + np.sum(nsky[ind_blue]) * nspaxel_x * nspaxel_y
        tot_source_counts_red = np.sum(nsource_cube[ind_red, :, :]) + np.sum(nsky[ind_red]) * nspaxel_x * nspaxel_y
        nstray_blue = tot_source_counts_blue * stray_frac / (4000 * 2000)
        nstray_red = tot_source_counts_red * stray_frac / (6000 * 3000)
        nstray = np.zeros(shape=nw, dtype=float)
        nstray[ind_blue] = nstray_blue
        nstray[ind_red] = nstray_red


        # make into 3d cube
        ndark_cube = repeat(ndark, 'h -> h ' + str(nspaxel_x) + ' ' + str(nspaxel_x))
        nreadout_cube = repeat(nreadout, 'h -> h ' + str(nspaxel_x) + ' ' + str(nspaxel_x))
        nsky_cube = repeat(nsky, 'h -> h ' + str(nspaxel_x) + ' ' + str(nspaxel_x))
        nstray_cube = repeat(nstray, 'h -> h ' + str(nspaxel_x) + ' ' + str(nspaxel_y))

        self.ccdspec_wave = ccdspec_wave
        self.ndark = ndark_cube
        self.nreadout = nreadout_cube
        self.nsky = nsky_cube
        self.nsource = nsource_cube
        self.nstray = nstray_cube


        # consider the binning in x-axis
        if instru['readout_xbin'] > 1:
            ccdspec_wave_xbin = np.array([])
            ndark_bin = np.array([])
            nreadout_bin = np.array([])
            nsky_bin = np.array([])
            nstray_bin = np.array([])
            i = 0
            xbin = instru['readout_xbin']
            while i + xbin < nw:
                ccdspec_wave_xbin = np.append(ccdspec_wave_xbin, np.mean(ccdspec_wave[i:i+xbin]))
                ndark_bin = np.append(ndark_bin, np.mean(ndark[i:i+xbin]))
                nreadout_bin = np.append(nreadout_bin, np.mean(nreadout[i:i+xbin]))
                nsky_bin = np.append(nsky_bin, np.sum(nsky[i:i + xbin]))
                nstray_bin = np.append(nstray_bin, np.sum(nstray[i:i + xbin]))
                i = i + xbin

            # bin the source counts
            nw_bin = len(ccdspec_wave_xbin)
            nsource_bin_cube = np.zeros(shape=(nw_bin, nspaxel_x, nspaxel_y), dtype=float)
            for i in range(nw_bin):
                j = i * xbin
                nsource_bin_cube[i, :, :] = np.sum(nsource_cube[j:j + xbin, :, :], axis=0)

            # make into 3d cube (xbinning)
            ndark_bin_cube = repeat(ndark_bin, 'h -> h ' + str(nspaxel_x) + ' ' + str(nspaxel_x))
            nreadout_bin_cube = repeat(nreadout_bin, 'h -> h ' + str(nspaxel_x) + ' ' + str(nspaxel_x))
            nsky_bin_cube = repeat(nsky_bin, 'h -> h ' + str(nspaxel_x) + ' ' + str(nspaxel_x))  # e/s
            nstray_bin_cube = repeat(nstray_bin, 'h -> h ' + str(nspaxel_x) + ' ' + str(nspaxel_y))

            self.ccdspec_wave = ccdspec_wave_xbin
            self.ndark = ndark_bin_cube
            self.nreadout = nreadout_bin_cube
            self.nsky = nsky_bin_cube
            self.nsource = nsource_bin_cube
            self.nstray = nstray_bin_cube



class calculation_snr(object):

    def __init__(self, config):

        signal_rate = DetectorCountsRate(config)
        counts = DetectorCounts(config, signal_rate)

        self.source = counts.nsource
        self.darknoise = np.sqrt(counts.ndark)
        self.readnoise = np.sqrt(counts.nreadout)
        self.skynoise = np.sqrt(counts.nsky)
        self.sourcenoise = np.sqrt(counts.nsource)
        self.straynoise = np.sqrt(counts.nstray)
        self.saturate_mask = (self.source > 65535)


        self.totnoise = np.sqrt(self.darknoise ** 2 + self.readnoise ** 2 +
                                self.skynoise ** 2 + self.sourcenoise ** 2 +
                                self.straynoise ** 2)
        self.sysnoise = np.sqrt(self.darknoise ** 2 + self.readnoise ** 2 +
                                self.skynoise ** 2 + self.straynoise ** 2)
        self.snr = counts.nsource / self.totnoise


        # model spectrum
        nw, nx, ny = counts.nsource.shape
        self.mockwave = counts.ccdspec_wave
        template_fluxcube = np.zeros(shape=(nw, nx, ny), dtype=float)
        if config['configuration']['readout_xbin'] > 1:
            for i in range(nx):
                for j in range(ny):
                    template_fluxcube[:, i, j] = np.interp(self.mockwave,
                                                  signal_rate.ccdspec_wave,
                                                  signal_rate.fluxcube[:, i, j])
        else:
            template_fluxcube = signal_rate.fluxcube

        self.mockflux = np.random.normal(template_fluxcube, scale=template_fluxcube / self.snr)
        self.mockerror = template_fluxcube / self.snr

        # model image
        self.img2d = signal_rate.mag2d

    def save(self, file):

        f = h5py.File(file, 'w')
        f.create_dataset('source', data=self.source)
        f.create_dataset('darknoise', data=self.darknoise)
        f.create_dataset('readnoise', data=self.readnoise)
        f.create_dataset('skynoise', data=self.skynoise)
        f.create_dataset('sourcenoise', data=self.sourcenoise)
        f.create_dataset('totnoise', data=self.totnoise)
        f.create_dataset('sysnoise', data=self.sysnoise)
        f.create_dataset('straynoise', data=self.straynoise)
        f.create_dataset('snr', data=self.snr)
        f.create_dataset('saturate_mask', data=self.saturate_mask)
        f.create_dataset('mockwave', data=self.mockwave)
        f.create_dataset('mockflux', data=self.mockflux)
        f.create_dataset('mockerror', data=self.mockerror)
        f.create_dataset('img2d', data=self.img2d)
        f.close()


    def plot(self):

        report = self

        fig, axs = plt.subplots(figsize=(12, 9))
        fig.subplots_adjust(hspace=0.35, wspace=0.45,
                            left=0.05, right=0.97, top=0.95, bottom=0.05)
        cm = plt.cm.get_cmap('jet_r')

        ax_recontructed_magmap = plt.subplot2grid((3, 4), (0, 0))
        im = ax_recontructed_magmap.imshow(self.img2d, origin='low', extent=(0, 30, 0, 30),
                                           cmap=cm, picker=True)
        ax_recontructed_magmap.set_xlabel('spaxel')
        ax_recontructed_magmap.set_ylabel('spaxel')
        ax_recontructed_magmap.set_title('reconstructed magmap')
        cbar_ax = fig.add_axes([0.25, 0.75, 0.01, 0.2])
        cbar = plt.colorbar(im, cax=cbar_ax, ticks=np.arange(20, 25, 2), orientation='vertical')
        cbar.set_label(r'mag/arcsec$^{2}$', labelpad=-1)
        cbar.ax.tick_params(labelsize=7)

        ax_recontructed_snr = plt.subplot2grid((3, 4), (1, 0))
        ind = np.argmin(abs(self.mockwave - 5000))
        im = ax_recontructed_snr.imshow(self.snr[ind, :, :], origin='low', extent=(0, 30, 0, 30),
                                        cmap=cm, picker=True)
        ax_recontructed_snr.set_xlabel('spaxel')
        ax_recontructed_snr.set_ylabel('spaxel')
        ax_recontructed_snr.set_title('reconstructed SNR@5000A')
        cbar_ax = fig.add_axes([0.25, 0.41, 0.01, 0.2])
        cbar = plt.colorbar(im, cax=cbar_ax, ticks=np.arange(0, 4, 1), orientation='vertical')
        cbar.set_label('SNR', labelpad=-1)
        cbar.ax.tick_params(labelsize=7)

        k = 0
        i = 15
        j = 15
        ax_snr = plt.subplot2grid((3, 3), (k, 1), colspan=2)
        ax_snr.plot(self.mockwave, self.snr[:, j, i], 'k-')
        ax_snr.annotate('surface brightness: \n' + '{:4.1f}'.format(self.img2d[j, i]) + r' mag/arcsec$^{2}$',
                        xy=(0.75, 0.8), xycoords='axes fraction')
        ax_snr.annotate('SNR@5000A: \n' + '{:4.1f}'.format(self.snr[ind, j, i]),
                        xy=(0.75, 0.65), xycoords='axes fraction')
        ax_snr.annotate('x: ' + str(i) + ', y: ' + str(j),
                        xy=(0.05, 0.78), xycoords='axes fraction')
        ax_snr.set_ylabel('SNR')
        ax_snr.set_xlabel('wavelength')

        ax_noises = plt.subplot2grid((3, 3), (k + 1, 1), colspan=2)
        ax_noises.plot(self.mockwave, self.sourcenoise[:, j, i], label='source_noise')
        ax_noises.plot(self.mockwave, self.skynoise[:, j, i], label='sky_noise')
        ax_noises.plot(self.mockwave, self.readnoise[:, j, i], label='readout_noise')
        ax_noises.plot(self.mockwave, self.darknoise[:, j, i], label='dark_noise')
        ax_noises.plot(self.mockwave, self.straynoise[:, j, i], label='stray_noise')
        ax_noises.legend(loc=1)
        ax_noises.set_ylabel('noise counts (e-/s)')

        ax_recontructed_spec = plt.subplot2grid((3, 3), (k + 2, 1), colspan=2)
        ax_recontructed_spec.plot(self.mockwave, self.mockflux[:, j, i], 'k-', label='mockflux')
        ax_recontructed_spec.plot(self.mockwave, self.mockerror[:, j, i], 'b-', label='mockerr')
        ax_recontructed_spec.set_ylabel(r'mock flux (erg/s/cm$^2$/A)')
        ax_recontructed_spec.set_xlabel('wavelength')


        def onpick(event):

            x = event.mouseevent.xdata
            y = event.mouseevent.ydata
            # print('show spec in x = ' + str(int(np.floor(x))) + ', y = '+ str(int(np.floor(y))))

            xind = int(np.floor(x))
            yind = int(np.floor(y))

            update(xind, yind)

        def update(xind, yind):

            k = 0
            i = xind
            j = yind
            ax_snr = plt.subplot2grid((3, 3), (k, 1), colspan=2)
            ax_snr.plot(self.mockwave, self.snr[:, j, i], 'k-')
            ax_snr.annotate('surface brightness: \n' + '{:4.1f}'.format(self.img2d[j, i]) + r' mag/arcsec$^{2}$',
                            xy=(0.75, 0.8), xycoords='axes fraction')
            ax_snr.annotate('SNR@5000A: \n' + '{:4.1f}'.format(self.snr[ind, j, i]),
                            xy=(0.75, 0.65), xycoords='axes fraction')
            ax_snr.annotate('x: ' + str(i) + ', y: ' + str(j),
                            xy=(0.05, 0.78), xycoords='axes fraction')
            ax_snr.set_ylabel('SNR')
            ax_snr.set_xlabel('wavelength')

            ax_noises = plt.subplot2grid((3, 3), (k + 1, 1), colspan=2)
            ax_noises.plot(self.mockwave, self.sourcenoise[:, j, i], label='source_noise')
            ax_noises.plot(self.mockwave, self.skynoise[:, j, i], label='sky_noise')
            ax_noises.plot(self.mockwave, self.readnoise[:, j, i], label='readout_noise')
            ax_noises.plot(self.mockwave, self.darknoise[:, j, i], label='dark_noise')
            ax_noises.plot(self.mockwave, self.straynoise[:, j, i], label='stray_noise')
            ax_noises.legend(loc=1)
            ax_noises.set_ylabel('noise counts (e-/s)')

            ax_recontructed_spec = plt.subplot2grid((3, 3), (k + 2, 1), colspan=2)
            ax_recontructed_spec.plot(self.mockwave, self.mockflux[:, j, i], 'k-', label='mockflux')
            ax_recontructed_spec.plot(self.mockwave, self.mockerror[:, j, i], 'b-', label='mockerr')
            ax_recontructed_spec.set_ylabel(r'mock flux (erg/s/cm$^2$/A)')
            ax_recontructed_spec.set_xlabel('wavelength')

            fig.canvas.draw()


        fig.canvas.mpl_connect('pick_event', onpick)

        plt.show()


class read_report(object):

    def __init__(self, file):

        f = h5py.File(file, 'r')
        list(f.keys())

        self.source = f['source'].value
        self.darknoise = f['darknoise'].value
        self.readnoise = f['readnoise'].value
        self.skynoise = f['skynoise'].value
        self.sourcenoise = f['sourcenoise'].value
        self.totnoise = f['totnoise'].value
        self.sysnoise = f['sysnoise'].value
        self.straynoise = f['straynoise'].value
        self.snr = f['snr'].value
        self.saturate_mask = f['saturate_mask'].value
        self.mockwave = f['mockwave'].value
        self.mockflux = f['mockflux'].value
        self.mockerror = f['mockerror'].value
        self.img2d = f['img2d'].value
        f.close()


# class calculation_limitmag(object):
#
#     def __init__(self, config):
#
#         instru = config['configuration']
#         targetsnr = config['targetsnr']
#
#         ccdspec_wave = np.arange(3500, 10000, 1.75555)
#         darkc = instru['dark']      # e/s/pix
#         rn = instru['readout_noise']   # e/pix
#         nsky = get_background(ccdspec_wave)
#         repn = config['repn']
#         obst = config['obst']
#         npixw = 2
#         specsampling = 1
#
#         coeff_s2 = (repn * obst) ** 2
#         coeff_s1 = -targetsnr ** 2 * (repn * obst)
#         coeff_s0 = -targetsnr ** 2 * (rn ** 2 * (repn * npixw * specsampling) +
#                                 (darkc * repn * obst * npixw * specsampling) +
#                                 (nsky * repn * obst))
#
#         source_rate = (-coeff_s1 + np.sqrt(coeff_s1 ** 2 - 4 * coeff_s2 * coeff_s0)) / (2 * coeff_s2)
#
#         throughtputwave, throughtputvalue = get_throughput(config)
#         flux = electron2flux(config, ccdspec_wave, source_rate,
#                              throughtputwave, throughtputvalue,
#                              fluxunit='erg/s/cm^2/A')
#
#         filter = read_filter(config['scene'][0]['normalization']['band'])
#         ind_filter = (ccdspec_wave >= filter.wavemin) & (ccdspec_wave <= filter.wavemax)
#         filter_wave = ccdspec_wave[ind_filter]
#         filter_flux = np.interp(filter_wave, filter.wave, filter.throughput)
#         self.limitmag = filter_mag(ccdspec_wave, flux, filter_wave, filter_flux, output='mag')



def perform_calculation(config, calculation_mode='exptime2snr'):


    if calculation_mode == 'exptime2snr':

        report = calculation_snr(config)

    # elif calculation_mode == 'snr2exptime':
    #
    #     report = calculation_exptime(config)
    #
    # elif calculation_mode == 'snr2limitmag':
    #
    #     report = calculation_limitmag(config)

    else:

        print('calculation_type need to be " exptime2snr | snr2exptime | snr2limitmag "')
        pdb.set_trace()


    return report




import os
import json
import pandas as pd
import numpy as np
import pdb

snpp_refdata = os.path.abspath(os.path.dirname(__file__)+'/../refdata') + '/'

def get_telescope_config():

    dict = {}
    dict['diameter'] = 200
    dict['obscure'] = 0.
    dict['coll_area'] = 31415.926535897932

    # out_file = open(snpp_refdata + 'csst/telescope/config.json', "w")
    # json.dump(dict, out_file, indent=2)
    # out_file.close()


# def get_telescope_config():
#
#     config_file = snpp_refdata + 'csst/telescope/config.json'
#     f = open(config_file, 'r')
#     dict = json.load(f)
#     f.close()
#     return dict


def get_instrument_config():

    dict = {}
    # user parameter
    dict['readout_xbin'] = 1
    dict['readout_ybin'] = 1
    dict['gain'] = 1                            # e/ADU
    dict['dark'] = 0.017                        # e/s/pix
    dict['readout_noise'] = 4.0                 # e/pix
    dict['QE'] = 1.0                            # e/pix
    dict['efficiency_file'] = 'IFU_throughput.dat'

    # hidden parameters
    dict['ccd_xsize'] = 1.755555            # A, delta_lambda_per_pixel
    dict['ccd_ysize'] = 0.1                 # arcsec, spatial axis
    dict['extract_aper_width'] = 2 * dict['readout_ybin']   # extract spectrum with aperture of 2 pixels
    dict['spaxel_xsize'] = 0.2              # arcsec
    dict['spaxel_ysize'] = dict['extract_aper_width'] * 0.1   # arcsec
    dict['fov_xsize'] = 6                 # arcsec
    dict['fov_ysize'] = 6                 # arcsec
    dict['wave_start'] = 3500
    dict['wave_end'] = 10000
    dict['wave_delta'] = dict['ccd_xsize'] * dict['readout_xbin']

    # out_file = open(snpp_refdata + "csst/ifs/config.json", "w")
    # json.dump(dict, out_file, indent=2)
    # out_file.close()
    return dict


# def get_instrument_config():
#
#     config_file = snpp_refdata + 'csst/ifs/config.json'
#     f = open(config_file, 'r')
#     dict = json.load(f)
#     f.close()
#
#     return dict




def get_throughput():

    # throughput_file = os.path.join(os.getenv('SNPP_refdata'), 'csst', 'ifs', 'IFU_throughput.dat')
    # throughput = pd.read_csv(throughput_file,
    #                          sep='\s+', skiprows=1, header=None, names=['nm', 'q'])
    # lambdaq = throughput.nm.values * 10  # nm to A
    # qifs = throughput.q.values  # ; throughput of the whole system,
    # # ;assuming the total throughput cannot reach the theory value, 0.3 is the upper limit.
    # qifs[qifs >= 0.3] = 0.3
    # qinput = 1.0                    # the throughput of the telescope
    # qtot = qifs * qinput            # *qe ;qtot of CSST already includes the CCD efficiency


    # throughput_file = config['configuration']['efficiency_file']
    cat = pd.read_csv(snpp_refdata + 'csst/ifs/IFU_throughput.dat',
                      comment='#', sep='\s+', header=None, names=['nm', 'q'])
    throughputwave = cat['nm'].values * 10  # nm to A
    throughputvalue = cat['q'].values       # energe fraction or photon fraction??

    return throughputwave, throughputvalue





def build_default_source():

    dict = {}
    dict['id'] = 1
    dict['geometry'] = {'type': 'flat', 'x_offset': 0.0, 'y_offset': 0.0, 'PA': 45.0,
                        'Re_major': 3., 'Re_minor': 1.5, 'sersic_n': 1,
                        'xwidth': 2., 'ywidth': 2.}
    dict['spectrum'] = {'name': 'SFgal_texp_FeH0_tau5_Ew10_AGN1.fits', 'redshift': 0.0, 'lines':[]}
    dict['normalization'] = {'value': 17.7, 'unit':'mag/arcsec^2', 'band': 'SDSS-g'}

    # out_file = open(snpp_refdata + "source/config.json", "w")
    # json.dump(dict, out_file, indent=2)
    # out_file.close()

    return dict


def build_default_scene():

    scene = [build_default_source()]
    return scene




def build_default_calc():

    calc = {}
    calc['configuration'] = get_instrument_config()
    calc['source'] = build_default_source()
    calc['background_type'] = 'from_msc'
    calc['background_level'] = 'from_msc'
    calc['repn'] = 20.
    calc['obst'] = 300.

    calc['targetsnr'] = 10      # only be used in calculate_type = 'limitmag',
                                # in this case, normalization value is invalid




    return calc









from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='ifs_etc',
    version='0.0.3',
    description='exposure time calculator',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='linlin',
    author_email='linlin@shao.ac.cn',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={"ifs_etc": ["refdata/sed/*.fits",
                              "refdata/normalization/filters/*.par",
                              "refdata/csst/background/*.dat",
                              "refdata/csst/ifs/*.dat",
                              "refdata/csst/telescope/*",
                              "refdata/source/*"]},
    exclude_package_data={"": ["README.md"]},
    python_requires='>=3',
)




import numpy as np
import pandas as pd
import os
import astropy.io.fits as fits
from scipy.integrate import simps
from einops import repeat
import pdb

from astropy.modeling.models import Sersic2D, Box2D

from .snpp import input_mag_model, read_filter

snpp_refdata = os.path.abspath(os.path.dirname(__file__)+'/../refdata') + '/'

def sersic(nx, ny, xcen, ycen, ue, r50, b2a, PA, n):

    # 如果xscale yscale不一致的话，需要重新regrid。

    x, y = np.meshgrid(np.arange(nx), np.arange(ny))

    mod = Sersic2D(amplitude = ue, r_eff = r50, n=n, x_0 = xcen, y_0 = ycen,
                        ellip = 1-b2a, theta = (PA+90)/180*np.pi)
    # amplitude = surface brightness at r_eff


    # import matplotlib.pyplot as plt
    # plt.imshow(mod(x, y), origin='low')
    # plt.show()
    # pdb.set_trace()

    return mod(x, y)


def flatbox(nx, ny, xcen, ycen, u0, xwidth, ywidth):

    x, y = np.meshgrid(np.arange(nx), np.arange(ny))
    mod = Box2D(amplitude=u0, x_0=xcen, y_0=ycen, x_width=xwidth, y_width=ywidth)

    return mod(x, y)

class read_template(object):

    def __init__(self, source):

        self.source = source
        self.line_definitions = self.source['lines']

        # snpp_path = '/Users/linlin/Pro/snpp_v1_05/refdata/'
        template_filename = snpp_refdata + 'sed/' + self.source['name']
        # filtera = snpp_path + 'normalization/filters/sdss_g0.par'
        # result = input_mag_model(self.source['normalization']['value'],
        #                          galtpl, filtera)  # scale the template to given brightness
        hdu = fits.open(template_filename)
        self.wave = hdu[1].data['wavelength']   # A
        self.flux = hdu[1].data['flux'] * 1e-12         # erg/s/A/cm2


class get_grid(object):

    def __init__(self, config):

        self.config = config['configuration']
        self.xspxsize = self.config['spaxel_xsize']
        self.yspxsize = self.config['spaxel_ysize']
        self.nx = int(self.config['fov_xsize'] / self.xspxsize)
        self.ny = int(self.config['fov_ysize'] / self.yspxsize)
        self.spaxel_area = self.xspxsize * self.yspxsize
        self.x, self.y = np.meshgrid(np.arange(self.nx), np.arange(self.ny))

        self.object = config['source']
        self.xcen = self.nx/2 + self.object['geometry']['x_offset']
        self.ycen = self.ny/2 + self.object['geometry']['y_offset']


def get_wavearr(config):

    wave_start = 3500
    wave_end = 10000
    wave_delta = config['configuration']['ccd_xsize']
    # wave_delta = config['configuration']['ccd_xsize'] * config['configuration']['readout_xbin']

    ccdspec_wave = np.arange(wave_start, wave_end, wave_delta)

    return ccdspec_wave


class read_filter(object):
    # load the filters
    # filterfile = filtera  # '../sdss_g0.par'
    # filterpath='./'
    # filterfile=filterpath+filtersel   # ;fluxfilter: max=1, min=0, no particular unit
    def __init__(self, band):

        filterfile = snpp_refdata + 'normalization/filters/sdss_g0.par'
        print('Filter File:', filterfile)

        band = pd.read_csv(filterfile, sep='\s+', header=None, comment='#')
        self.wave = band[0].values  # A
        self.throughput = band[1].values  # vaccum_pass
        self.wavemin = self.wave[0]
        self.wavemax = self.wave[-1]

        # find the central wavelength, effective wavelength, and FWHM of the given filter
        filtermid = (self.wavemax - self.wavemin) * 0.5  # A, central wavelength
        dwave = self.wave[1:] - self.wave[:-1]
        self.waveeff = np.nansum(dwave * self.wave[1:] * self.throughput[1:]) / \
                       np.nansum(dwave * self.throughput[1:])
                                                                            # A, effective wavelength
        rmax = np.max(self.throughput)
        nnn = np.where(self.throughput > 0.5 * rmax)[0]
        self.FWHMmin = self.wave[nnn[0]]      # wave range at FWHM
        self.FWHMmax = self.wave[nnn[-1]]
        self.wavefwhm = self.FWHMmax - self.FWHMmin  # A, FWHM


def filter_mag(objwave, objflux, filterwave, filterthroughtput, output='mag'):
    '''
    :param objwave: unit as A
    :param objflux: unit as erg/s/cm^2/A
    :param filterwave: unit as A
    :param filterthroughtput: unit as detector signal per photon (use vaccum_pass for space telescope,
                              otherwise select a given airmass)
    :return: AB magnitude in this band
    '''

    # resample the throughtput to objwave
    ind = (objwave >= np.min(filterwave)) & (objwave <= np.max(filterwave))
    wavetmp = objwave[ind]
    phot_frac = np.interp(wavetmp, filterwave, filterthroughtput)  # phot fraction (/s/A/cm^2?)
    # convert to energy fraction
    # E_frac = E0/hv * N_frac / E0 ~ N_frac/nu ~ N_frac * lambda
    energy_frac = phot_frac * wavetmp  # convert to energy fraction

    # convert the objflux to erg/s/cm^2/Hz
    c = 3e18    # A/s
    objflux_hz = objflux[ind] * (objwave[ind]**2) / c    # erg/s/scm^2/Hz

    from scipy.integrate import simps
    integrate_flux = simps(objflux_hz * energy_frac, c/wavetmp) / simps(energy_frac, c/wavetmp)

    if output == 'mag':
        mag = -2.5 * np.log10(integrate_flux) - 48.6
        return mag
    elif output == 'flux':
        return integrate_flux
    else:
        print('Error: output need to be "mag" or "flux". ')
        return np.nan


def normalized(template_wave, template_flux, config):

    grid = get_grid(config)

    source = config['source']

    normalize = source['normalization']

    filter = read_filter(normalize['band'])

    ind_filter = (template_wave >= filter.wavemin) & (template_wave <= filter.wavemax)
    filter_wave = template_wave[ind_filter]
    filter_flux = np.interp(filter_wave, filter.wave, filter.throughput)
    filter_constant = simps(filter_flux * filter_wave, filter_wave)

    template_constant = simps(filter_flux * template_wave[ind_filter] * template_flux[ind_filter],
                              template_wave[ind_filter])

    u0 = normalize['value']
    u0 = 10**((u0 + 48.6)/(-2.5))         # target flux in erg/s/cm^2/Hz unit
    u0 = u0 * 3e18 / filter.waveeff**2    # erg/s/cm^2/A

    geometry = source['geometry']
    if geometry['type'] == 'Sersic':

        xcen = grid.nx / 2 + geometry['x_offset']
        ycen = grid.ny / 2 + geometry['y_offset']
        rmaj = geometry['Re_major'] / grid.xspxsize
        rmin = geometry['Re_minor'] / grid.xspxsize

        scaled_img = sersic(grid.nx, grid.ny, xcen, ycen,
                 u0, rmaj, rmin/rmaj, geometry['PA'], geometry['sersic_n'])

    elif geometry['type'] == 'flat':

        xcen = grid.nx / 2 + geometry['x_offset']
        ycen = grid.ny / 2 + geometry['y_offset']

        scaled_img = flatbox(grid.nx, grid.ny, xcen, ycen,
                        u0, geometry['xwidth'], geometry['ywidth'])

    else:
        print('geometry should be one of these: ("Sersic" | "flat").')
        pdb.set_trace()

    factor = scaled_img * filter_constant / template_constant

    norm_wave = repeat(template_wave, 'h -> h w c', w=grid.ny, c=grid.nx)
    tmp_template_flux = repeat(template_flux, 'h -> h w c', w=grid.ny, c=grid.nx)
    tmp_factor = repeat(factor, 'w c -> h w c', h=len(template_wave))
    norm_flux = tmp_template_flux * tmp_factor


    # check the normalized magnitude should be normalized.value - 2.5*log(0.04)
    mag2d = np.zeros(shape=(grid.ny, grid.nx), dtype=float)
    for j in range(grid.ny):
        for i in range(grid.nx):
            mag2d[j, i] = filter_mag(norm_wave[:, j, i], norm_flux[:, j, i],
                                   filter_wave, filter_flux, output='mag')


    return norm_wave, norm_flux, mag2d


class ModelCube(object):

    def __init__(self, config):

        self.ccdspec_wave = get_wavearr(config)
        self.ccdspec_nw = len(self.ccdspec_wave)

        grid = get_grid(config)

        object = config['source']['spectrum']

        template = read_template(object)
        template_wave_interp = self.ccdspec_wave
        template_flux_interp = np.interp(template_wave_interp, template.wave, template.flux)

        # self.ccdspec_flux = np.zeros(shape=(self.ccdspec_nw, grid.ny, grid.nx), dtype=np.float32)

        self.wavecube, self.fluxcube, self.mag2d = normalized(template_wave_interp, template_flux_interp, config)


    # def export_to_fits(self, fitsfile='ModelCube'):
    #



# class ConvolvedCube(object):

